/* 
 * Huffman coding
 *
 * 15-122 Principles of Imperative Computation
 */

#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <assert.h>

#include "lib/contracts.h"
#include "lib/xalloc.h"
#include "lib/heaps.h"

#include "freqtable.h"
#include "huffman.h"

struct htree_node {
  char value;			/* '\0' except at leaves */
  int frequency;
  struct htree_node *left;
  struct htree_node *right;
};

/* in huffman.h */
/* typedef struct htree_node htree; */

void error(char *msg) {
  fprintf(stderr, "%s\n", msg);
  exit(1);
}

/**************************************/
/* Task 5a                            */
/* Checking data structure invariants */
/**************************************/

/*************************************************/
/* FILL IN THESE FUNCTIONS, STARTING FROM TASK 1 */
/*************************************************/

bool is_htree0(htree *H);
bool is_htree2(htree *H);
bool is_htree(htree *H);

/************************************************/
/* Task 5b                                      */
/* Building Huffman trees from frequency tables */
/************************************************/

/*************************************************/
/* FILL IN THESE FUNCTIONS, STARTING FROM TASK 2 */
/*************************************************/

/* for libs/heaps.c */
int elem_priority(elem e);

htree *build_htree(int *table);

/*** new in C ***/
void htree_free(htree *H) {
  (void) H;  /* avoids 'unused parameter' warning until code is written */

  /*******************************************/
  /* FILL IN THIS CODE, STARTING FROM TASK 2 */
  /*******************************************/

  return;
}

/*******************************************/
/* Bonus Task                              */
/* Building code tables from Huffman trees */
/*******************************************/

bool is_bitstring(char *s)
{
  if (s == NULL) return false;
  int i = 0;
  char c = s[i];
  while (c != '\0') {
    if (!(c == '0' || c == '1')) return false;
    i++;
    c = s[i];
  }
  return true;
}

bool is_codetable(char **C) {
  if (C == NULL) return false;
  for (int i = 0; i < 128; i++)
    if (!(C[i] == NULL || is_bitstring(C[i]))) return false;
  return true;
}

char **htree_to_codetable(htree *H) {
  (void) H;  /* avoids 'unused parameter' warning until code is written */

  /************************************/
  /* FILL IN THIS CODE FOR BONUS ONLY */
  /************************************/

  return NULL;
}

void codetable_free(char **table) {
  (void) table;  /* avoids 'unused parameter' warning until code is written */

  /************************************/
  /* FILL IN THIS CODE FOR BONUS ONLY */
  /************************************/

}

char *encode(htree *H, char *input)
{
  (void) H;      /* avoids 'unused parameter' warning until code is written */
  (void) input;  /* avoids 'unused parameter' warning until code is written */

  /************************************/
  /* FILL IN THIS CODE FOR BONUS ONLY */
  /************************************/

  return NULL;
}

/* Decoding bitstrings */
/* Provided by the course staff! */

char *decode(htree *H, char *bits) {
  REQUIRES(is_htree2(H) && bits != NULL);
  int len = strlen(bits);

  /* start by determining length of decoded msg */
  int count = 0;
  htree *node = H;
  int k = 0;

  while (k < len)
    //@loop_invariant 0 <= k && k <= len;
    //@loop_invariant is_htree2(node);
    {
      if (bits[k] == '0') {
	node = node->left;
	k++;
      } else if (bits[k] == '1') {
	node = node->right;
	k++;
      } else if (bits[k] == '\0') {
	error("Unexpected end of string");
      } else {
	error("Illegal character in input");
      }
      if (node->value != '\0') {
	ASSERT(is_htree0(node));
	count++;
	node = H;
      }
    }
  ASSERT(is_bitstring(bits));

  if (!(node == H && k == len)) {
    error("Early string termination");
  }
  ASSERT(node == H);

  char *decoded_msg = xcalloc(count+1, sizeof(char));
  int j = 0;
  k = 0;

  while (k < len)
    //@loop_invariant 0 <= k && k <= len;
    //@loop_invariant is_htree2(node);
    {
      if (bits[k] == '0') {
	node = node->left;
	k++;
      } else if (bits[k] == '1') {
	node = node->right;
	k++;
      }
      if (node->value != '\0') {
	decoded_msg[j] = node->value;
	j++;
	node = H;
      }
    }
  decoded_msg[j] = '\0';
  return decoded_msg;
}
